/*
 * parser.* - configuration file parser
 * Copyright (C) 2005-2006 Shobbo, urug
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <unistd.h>	/* read	   */
#include <stdlib.h>	/* free	   */
#include <stdio.h>	/* printf  */
#include <string.h>	/* strncmp */
#include <strings.h>	/* index   */
#include <fcntl.h>      /* open    */
#include <stdarg.h>	/* va_start*/
#include "parser.h"


int      cur_line = 1;		/* Parser location. */
#define	 AERROR(a,x)    	{ error(0, 0, NULL, a, NULL); return x; } /* Simple error alias. */


static void	error( int pos, int size, char *string, ... );

static boolean	p_variable_a( STRING *string, VAR **root );
static boolean	p_include_a( STRING *string, INCLUDE **root );

static boolean	p_parse_include_args( STRING *string, struct inc_args *arguments );
static boolean	p_exclude( INCLUDE *root );

static boolean	p_replace_vars( STRING *string, VAR *root );
static short	p_writestring( STRING *string, VAR *root, int fd );
static boolean	p_decompose( STRING *string );
static void	p_free_var( VAR *root );

static boolean	_p_fill_perms( struct inc_perms *perms, char *start, int size );
static boolean	_p_change( struct inc_args *conf, boolean brace, char *start, int size );
static boolean	_p_check( struct inc_args *conf, boolean brace, char *start, int size );
static boolean	_p_deep( struct inc_args *conf, boolean brace, char *start, int size );
static boolean	_p_cur( struct inc_args *conf, boolean brace, char *start, int size );
static boolean	_p_sub( struct inc_args *conf, boolean brace, char *start, int size );



/*
** Function prints errors. List of arguments must be ended with 
** NULL. Use pos/size/string if you want to specify string which
** dont contain '\0' at the end. (pos - position of string).
*/
static void
error( int pos, int size, char *string, ... )
{
    va_list	ap;
    int		counter = 1;
    char	*ep, *ptr;


    /* Initialize ap. */
    va_start(ap, string);

    /* Print line, start printing errors. */
    fprintf(stdout, "critical error at line(%d): ", cur_line);
    do
    {
        ep = va_arg(ap, char *);

        /* Print 'string' at 'pos' */
        if( string != NULL && pos == counter )
            for( ptr = string; ptr < string + size; ptr++)
                  printf("%c", *ptr);

        /* Print next argument. */
        if( ep != NULL ) fprintf(stdout, "%s", ep);

        /* Increment counter. */
        counter++;
    } 
    while( ep != NULL );

    /* Print newline and end */
    printf(".\n");
    va_end(ap);
}



/*
** Function starts parser.
*/
boolean
p_init( INCLUDE **inc, char *config )
{
    STRING	string;
    VAR		*var_root = NULL;
    boolean	ret = TRUE;
    short	status;
    int		fd;


    /* Open config file descriptor, print error if file doesn't exist. */
    if( (fd = open(config, O_RDONLY)) == -1 )
    {
          perror(config);
          return FALSE;
    }


#ifndef NO_DEBUG
    printf("\n*-*-*-*-*-*-*-*-*-*-\n"\
           "| Reading data from config file descriptor.\n"\
           "-*-*-*-*-*-*-*-*-*-*-\n");
#endif    

    /* Read config file until get error or EOF. */
    while( (status = p_writestring(&string, var_root, fd)) > 0 )
    {
#ifndef NO_DEBUG
	printf("|%s|\n", string.buffer);
#endif

        /* Add variable or rule. */
	switch(string.type)
        {
            case i_VAR: ret = p_variable_a(&string, &var_root); break;
            case i_INC: ret = p_include_a(&string, inc); break;
        }

	/* If one of above functions returned error, stop parser. */
	if( ret == FALSE ) break;
    }

    /* Create exclude list. */
    if( status != -1 && ret != FALSE )
        if( p_exclude(*inc) == FALSE )
            status = FALSE;

    /* Free memory from VAR root. */
    p_free_var(var_root);

    /* Close config file descriptor. */
    close(fd);

    if( status == -1 || ret == FALSE )
        return FALSE;
    else
        return TRUE;
}



/*
** Read signs from config file, until get semicolon. Returns 0 on EOF, and
** -1 on error.
**
** Warning: function skips all:
** 	* first space signs in string;
**	* duplicated space signs;
**	* last space signs;
**	* duplicated '/' signs;
*/
static short
p_writestring( STRING *string, VAR *root, int fd )
{
    int		status;
    boolean	skip_comments = FALSE;
    char        buf, last_sign, 
                *cur_sign = string->buffer;


    last_sign = *cur_sign = '\0';
    do
    {
        /* If read returned EOF, string->buffer should be empty. */
        if( (status = read( fd, &buf, 1)) == 0 )
        {
	    if( string->buffer[0] != '\0' )
	        AERROR("missing semicolon", -1);

            return 0;
        }

	/* Skip all comments. */
	if( skip_comments == TRUE && buf != '\n' )
	    continue;

        /* Check sign in buf. */
        switch( buf )
        {
            case '\n':
                  skip_comments = FALSE;
                  cur_line++;
                  break;

            case ';':
                  /* Skip last spaces. */
                  while( cur_sign > string->buffer && *--cur_sign == ' ' );
                  *++cur_sign = '\0'; /* Add NUL at end. */

                  /* Replace vars, and run p_decompose on string. */
                  if( string->buffer[0] != '\0' )
                  { 
                      if( p_replace_vars(string, root) == FALSE ) return -1;
                      if( p_decompose(string) == FALSE ) return -1;

                      return 1;
                  }
                  else
                      AERROR("lonely semicolon", -1);
                  
            case '#':
	          skip_comments = TRUE;
	          break;

            default:
                if( buf == '\t' )
                    buf = ' ';

                /* Skip duplicated spaces and '/', skip space if is first sign in rule. */
                if( (buf == '/' && last_sign == '/') || ( buf == ' ' && last_sign == ' ') || ( buf == ' ' && cur_sign == string->buffer ) )
                    break;

                *cur_sign++ = last_sign = buf;
        }
    }
    while( status > 0 && cur_sign < string->buffer + STRING_MAX_SIZE );


    /* Print error if rule is too long. */
    if( cur_sign >= string->buffer + STRING_MAX_SIZE )
        AERROR("rule is too long", -1);

    /* Return error. */
    perror("read");
    return -1;
}



/*
** Function writes to struct INCLUDE rule options, and filenames.
** Warning: function skips / if is last sign before list and first
** in list.
*/
static boolean
p_include_a( STRING *string, INCLUDE **root )
{
    FILES		**files;
    static INCLUDE	**node;
    boolean		list_eof = TRUE;
    short		name_size, lele_size;
    char		*filename_first = string->name[0];
    char		*filename_last = string->name[1];
    char 		*list_start_cut = NULL,
                        *list_stop_cut = NULL;


    /* Go to next element. */
    if( node == NULL )
        node = root;
    else
        node = &(*node)->next;

    /* alloc memory for node, and clear all values. */
    *node = alloc(sizeof(INCLUDE));
    files = &(*node)->include;
    (*node)->next = NULL;
    (*node)->include = NULL;
    (*node)->exclude = NULL;
    (*node)->line_in_config = cur_line;

    /* Parse arguments list. */
    if( p_parse_include_args(string, &(*node)->inc_args) == FALSE )
        return FALSE;

    /* Append all elements from list. */
    if( string->list[0] != NULL )
    {
        list_eof = FALSE;
        list_start_cut = string->list[0];
    }

    /* Save sizes. */
    name_size = filename_last - filename_first;
    lele_size = 0;


    do
    {
        /* Check list. */
        if( string->list[0] != NULL )
        {
            /* string->list[1] is never last element in rule, so we dont
               need to check NULL string. */
            list_stop_cut = index(list_start_cut, ' ');
            if( list_stop_cut >= string->list[1] )
            {
                list_stop_cut = string->list[1];
                list_eof = TRUE;
            }

            /* Delete slash if is last sign before list and first in list. */ 
            if( filename_first[name_size-1] == '/' && *list_start_cut == '/' )
                list_start_cut++;
        
            /* Delete all slashes at end of filename. */
            while( list_stop_cut - list_start_cut >= 1 && *--list_stop_cut == '/' );
            list_stop_cut++;
        
            /* Save list element size. */
            lele_size = list_stop_cut - list_start_cut;
        }
        
        *files = alloc( sizeof(FILES) );
        (*files)->filename = alloc( (name_size + lele_size + 1) * sizeof(char));
        (*files)->filename[0] = '\0';
        strncat((*files)->filename, filename_first, name_size);

        /* If user declared list, append list element to filename. */
        if( string->list[0] != NULL )
        {
            strncat((*files)->filename, list_start_cut, lele_size );

            /* Ommit all last (ommited) slashes. */
            while( *list_stop_cut++ == '/' );
            list_start_cut = list_stop_cut;
        }

        (*files)->next = NULL;
        files = &(*files)->next;
    }
    while( list_eof != TRUE );


    return TRUE;
}



/*
** Add variable from string to struct VAR.
*/
static boolean
p_variable_a( STRING *string, VAR **root )
{
    VAR 	**node = root;
    char 	*name_first_sign = string->name[0];
    char	*name_last_sign = string->name[1];
    int		name_size = name_last_sign - name_first_sign;


    /* Check for variable with same name. */
    for( ; *node != NULL ; node = &(*node)->next )
        if( strlen((*node)->name) == name_size )
            if( strncmp((*node)->name, name_first_sign, name_size) == 0 ) 
            {
	        error(2, name_size, name_first_sign, "duplicated variable: '","'", NULL);
                return FALSE;
            }

    /* Allocate memory for new struct. */
    *node = alloc( sizeof(VAR) );

    /* Copy variable name. */
    (*node)->name = alloc((name_size + 1) * sizeof(char));
    (*node)->name[0] = '\0';
    strncat((*node)->name, name_first_sign, name_size);

    /* Copy variable content. */
    ++name_last_sign; /* skip space. */
    (*node)->contents = alloc((strlen(name_last_sign) + 1 ) * sizeof(char));
    strcpy((*node)->contents, name_last_sign); /* I use strcpy instead of strcat, because */
                                               /* string has \0 at the end. */
    (*node)->next = NULL;

    return TRUE;
}



/*
** Replace all variable names $(name) with real contents.
** Warning: 
**	Function skips '/' if is last sign before variable name and first in
**	variable, and if is last sign in variable name and first after variable. 
**	(I dont want duplicated '/')
*/
static boolean
p_replace_vars( STRING *string, VAR *root )
{
    VAR 	*node = root;
    char	*backup, *var_content;
    char 	*start_cut, *stop_cut, 		 /* Where var starts and stops  $ and ) */
                *var_name_start, *var_name_end;  /* First and last sign of var name. */
    int		 var_name_size, var_content_size;


    /* Find beginning and end of variable name. */ 
    start_cut = string->buffer;
    do
    {
	if( (var_name_start = start_cut = index(start_cut, '$') ) == NULL )
	    return TRUE;
        else
	{
	      /* If found lonely dolar sign, skip it and start again. */
              if( *++var_name_start != '(' )
              {
                  start_cut = var_name_start;
                  continue;
              }
              else
                  var_name_start++;	/* If not, skip ( */

              /* Search closing brace. */
              if( (var_name_end = index(var_name_start,')')) == NULL )
                  AERROR("missing brace in variable", FALSE)
	}

	/* Save variable name size and place where variable ends (After ) sign) */
	var_name_size = var_name_end - var_name_start;
        stop_cut = var_name_end + 1;
        break;
    }
    while( 1 );


    /* Search for variable name in struct VAR. */
    for( ; node != NULL ; node = node->next )
	if( strlen(node->name) == var_name_size )
	    if( strncmp( node->name, var_name_start, var_name_size ) == 0 ) 
		break;
    if( node == NULL )
    {
	error(2, var_name_size, var_name_start, "unknown variable: '", "'", NULL);
	return FALSE;
    }

    /* Save variable content and size. */
    var_content = node->contents;
    var_content_size = strlen(node->contents);

    /* Skip / if is last sign before variable and first in variable. */
    --start_cut; /* start_cut is pointing to $ */
    if( *start_cut++ == '/' && *var_content == '/' )
    {
        var_content_size--;	/* Change size. */
        var_content++;		/* Skip first sign. */
    }

    /* Skip / if is last sign in variable and first after variable. */
    if( var_content[var_content_size - 1] == '/' && *stop_cut == '/' )
        var_content_size--;

    /* Check new string length. */
    if( (strlen(string->buffer) + var_content_size - var_name_size) >= STRING_MAX_SIZE )
	AERROR("rule is too long", FALSE);

    /* Copy everything after variable from string to backup. */
    backup = alloc((strlen(stop_cut) + 1 ) * sizeof(char));
    strcpy(backup, stop_cut);

    /* Copy variable contents. */
    *start_cut = '\0'; /* Write NUL at $ sign (from $( ) ) */
    strncat(start_cut, var_content, var_content_size);

    /* Copy backup (All after variable name) */
    strcat(start_cut, backup);

    /* Free backup memory. */
    free(backup);

    /* Start recurrence. */
    if( p_replace_vars(string, root) == FALSE)
        return FALSE;

    return TRUE;
}



/*
** Fills STRING struct.
*/
static boolean
p_decompose( STRING *string )
{
    char	*start_cut, *stop_cut, *temp;
    int		size;


    /* Get pointer to first space (Which separate name with dest) */
    if( (start_cut = index(string->buffer, ' ')) == NULL )
        AERROR("too few args in rule", FALSE);
    size = start_cut - string->buffer;

    /* Save instruction type. */
    if( strncmp(string->buffer, "v&", size) == 0 )
        string->type = i_VAR;
    else if( strncmp(string->buffer, "i&", size) == 0 )
        string->type = i_INC;
    else
    {
        error(2, size, string->buffer, "unknown instruction: '", "'", NULL);
        return FALSE;
    }


    /* Save destination (variable or file/directory name) */
    start_cut++; /* Skip space. */
    if( (stop_cut = index(start_cut, ' ')) == NULL )
        AERROR("too few args in rule", FALSE)
    else
    {
        /* Save positions. */
        string->name[0] = start_cut;
        string->name[1] = stop_cut;
        string->list[0] = NULL;
        
        if( string->type == i_INC )
        {
            /* Search for list. */
            for( temp = string->name[0] ; temp != string->name[1] ; temp++ )
                if( *temp == '$' )
                {
                    if( *++temp == '[' )
                    {
                        string->name[1] = --temp; 	/* Mark new name ending. (Before [) */
                        temp += 2; 			/* Skip $[ */
                        if( *temp == ' ') temp++;	/* Skip first space in list. */

                        /* Save pointer to beggining and ending of list. */
                        string->list[0] = temp;
                        if( (temp = index(temp, ']')) == NULL )
                            AERROR("missing brace in list", FALSE);

                        /* Mark stop_cut after ] */
                        stop_cut = temp+1;

                        /* Skip last space in list, mark ending of list. */ 
                        if( string->list[0] != temp && *--temp != ' ') temp++;
                        string->list[1] = temp;

                        /* Check length. */
                        if( string->list[1] - string->list[0] == 0 )
                            AERROR("empty list", FALSE);

                        break;	/* Break from loop */
                    }
                    else
                        temp--;
                }

            /* Delete ending slash in filename */
            temp = string->name[1];
            if( *--temp != '/' ) temp++;

            /* Leave slash if filename is lonely slash or list element. */
            if( string->list[0] != NULL || temp - start_cut == 0 )
                if( *temp == '/' ) 
                    temp++;

            /* Write new name ending. */
            string->name[1] = temp;
        }
    }

    /* Check if stop_cut is space */
    if( *stop_cut == '\0')
        AERROR("too few args in rule", FALSE)
    if( *stop_cut == ' ') stop_cut++;	/* Skip first space. */

    /* Save first and last sign of arguments list. */
    string->args[0] = stop_cut;
    string->args[1] = &string->buffer[strlen(string->buffer)];

    return TRUE;
}



/*
** Function excludes files and directories from include (Only for sub option).
** It excludes files/directories which are deeper than rule with sub.
*/
static boolean
p_exclude( INCLUDE *root )
{
    INCLUDE	*node, *search;
    EXCLUDE	**exclude;
    FILES	*files_node, *files_search;
    char	*compare1, *compare2, *temp;
    int		compare1_size, deep;


    /* Search for rules with sub arg. */
    for( node = root;  node != NULL; node = node->next )
        if( node->inc_args.sub == TRUE )
            for( files_node = node->include; files_node != NULL; files_node = files_node->next )
            {
                compare1 = files_node->filename;
                compare1_size = strlen(compare1);

                /* Search for end of exclude list. */
                for( exclude = &node->exclude; *exclude != NULL; exclude = &(*exclude)->next );

                /* Now compare compare1 element with all other rules. */
                for( search = root; search != NULL; search = search->next )
                    for( files_search = search->include; files_search != NULL; files_search = files_search->next )
                    {
                        /* Compare compare2 with compare1 (but use size of compare1) */
                        compare2 = files_search->filename;
                        if( compare1 != compare2 && strncmp(compare1, compare2, compare1_size) == 0 )
                        {
                            /* Print error if file is same as in node */
                            if( compare1_size == strlen(compare2) )
                            {
                                printf("critical error: duplicated file or directory in include list: \'%s\'\n", compare2);
                                return FALSE;
                            }

                            /* Count deep from place, where compare1 ends. */
                            for( deep = 0, temp = &compare2[compare1_size]; *temp != '\0'; temp++ )
                                if( *temp == '/' )
                                    deep++;

                            /* If destination is lonely / add one to deep. */
                            if( compare1_size == 1 && *compare1 == '/' )
                                deep++;

                            /* To exclude this file its deep minus node deep
                               should be zero or bigger. */
                            if( node->inc_args.deep != 0 )
                                if( (deep - node->inc_args.deep) > 0 )
                                    break;

                            /* Alloc memory. */
                            *exclude = alloc(sizeof(EXCLUDE));
                            (*exclude)->filename = compare2;
                            (*exclude)->next = NULL;

                            /* Clear params. */
                            (*exclude)->exc_args.cur = FALSE;
                            (*exclude)->exc_args.sub = FALSE;
                            (*exclude)->exc_args.deep = 1;
                            
                            /* Write params. */
                            if( search->inc_args.cur == TRUE ) (*exclude)->exc_args.cur = TRUE;
                            if( search->inc_args.sub == TRUE )
                            {
                                (*exclude)->exc_args.sub = TRUE;
                                (*exclude)->exc_args.deep = search->inc_args.deep;
                            }
                            
                            /* Go to next element. */
                            exclude = &(*exclude)->next;
                        }
                    }
            }
            
    return TRUE;  
}



/*
** Parse options and fill config
*/
static boolean
p_parse_include_args( STRING *string, struct inc_args *arguments )
{
    char	*start_name, *end_name;
    char	*start_opts, *end_opts;
    char	*end_arg;
    int	 	size_name, size_opts=0;
    boolean	brace;

    struct sparse
    {
          char		*cmd;
          int		size;
          boolean	(*func)(struct inc_args *, boolean, char *, int);
    } *cmds, sparse[] = { 
                  { "change", 6, _p_change },
                  { "check",  5, _p_check },
                  { "deep",   4, _p_deep },
                  { "cur",    3, _p_cur },
                  { "sub",    3, _p_sub },
                  { NULL, 0, NULL } };

    /* Clear args. */
    arguments->check = FALSE;
    arguments->change = FALSE;
    arguments->cur = FALSE;
    arguments->sub = FALSE;
    arguments->deep = 1;

    /* Start parsing arguments list. */
    for( start_opts = start_name = string->args[0]; start_name < string->args[1]; start_name = ++end_arg )
    {
          /* Find ending of argument name. */
          for( end_name = start_name; *end_name != '\0' && *end_name != '(' && *end_name != ' ' && end_name != string->args[1]; end_name++ );

          /* Save arg size. */ 
          size_name = end_name - start_name;

          /* Skip ( or ' ' */
          if( *(end_arg = end_name) == '(' ) end_arg++;
 
          /* Get argument options. */
          if( *end_name == '(' )
          {
                brace = TRUE;
                
                /* Search for closing brace. */
                for( end_opts = start_opts = end_name; *end_opts != ')' && *end_opts != '\0' &&\
                     end_opts != string->args[1]; end_opts++ );
                if( *end_opts != ')' )
                {
                    error(2, size_name, start_name, "missing closing brace in argument: '","'", NULL);
                    return FALSE;
                }

                /* Skip brace ). */
                end_arg = end_opts+1;

                /* Skip ( */
                start_opts++;

                /* Skip first and last space in options list. */ 
                if( *start_opts == ' ') start_opts++;
                if( start_opts != end_opts && *--end_opts != ' ') end_opts++;

                /* Save size. */
                size_opts = end_opts - start_opts;

                /* Check size of argument options. */
                if( size_opts == 0 )
                    AERROR("empty options list", FALSE)
                else if( size_opts >= OPTS_MAX_SIZE )
                    AERROR("options in argument too long", FALSE)
          }
          else
                brace = FALSE;

          /* Run minor parser functions. */
          for( cmds = &sparse[0]; cmds->cmd != NULL; cmds++ )
              if( size_name == cmds->size && strncmp(cmds->cmd, start_name, size_name) == 0 )
              {
                  if( cmds->func(arguments, brace, start_opts, size_opts) == FALSE )
                      return FALSE;
                  break;
              }
          if( cmds->cmd == NULL )
          {
              error(2, size_name, start_name, "unknown argument: '", "'", FALSE);
              return FALSE;
          }
    }

    /* Check if sub or cur is declared. Save uid, gid and mode. */
    if ( arguments->sub == FALSE && arguments->cur == FALSE )
        AERROR("you must set sub or cur (or both)", FALSE)

    /* Check if sub or cur is declared. Save uid, gid and mode. */
    if ( arguments->check == FALSE && arguments->change == FALSE )
        AERROR("you must set change or check (or both)", FALSE)

    return TRUE;
}



/*
** Fill struct perms *perms with owner, group and mode.
*/
static boolean
_p_fill_perms( struct inc_perms *perms, char *start, int size )
{
    char	buf[5];
    char	*ptr, *start_cut, *stop_cut;
    int		counter;


    /* Check if user defined 3 options in this argument. */
    for( counter = 0, ptr = start ; *ptr != '\0' && ptr < start + size; ptr++ )
        if( *ptr == ' ' )
            counter++;
    if( counter != 2 )
        AERROR("cur and sub need 3 options (perms group owner)", FALSE);

    /* Write permissions. */
    start_cut = start;
    stop_cut = index(start_cut, ' ');
    if( stop_cut - start_cut > 4 || stop_cut - start_cut < 4 )
        AERROR("Wrong access permissions (should be a 4-char-len list of rights eg. 0750)", FALSE);
    buf[0] = '\0';
    strncat(buf, start_cut, stop_cut - start_cut);
    perms->mode = strtol(buf, 0, 8);

    /* Write user. */
    start_cut = ++stop_cut;	/* Skip space. */
    stop_cut = index(start_cut,' ');
    perms->user = alloc( (stop_cut - start_cut + 1) * sizeof(char));
    perms->user[0] = '\0';
    strncat(perms->user, start_cut, stop_cut - start_cut );

    /* Write group. */
    start_cut = ++stop_cut;	/* Skip space. */
    stop_cut = start + size;
    perms->group = alloc( (stop_cut - start_cut + 1) * sizeof(char));
    perms->group[0] = '\0';
    strncat(perms->group, start_cut, stop_cut - start_cut );

    return TRUE;
}



/*
** Switch ON args.change.
*/
static boolean
_p_change( struct inc_args *conf, boolean brace, char *start, int size )
{
    if( brace == TRUE ) AERROR("argument 'change' takes no option", FALSE);
    conf->change = TRUE;
    return TRUE;
}



/*
** Switch ON args.check
*/
static boolean
_p_check( struct inc_args *conf, boolean brace, char *start, int size )
{
    if( brace == TRUE ) AERROR("argument 'check' takes no option", FALSE);
    conf->check = TRUE;
    return TRUE;
}



/*
** Switch ON args.deep() and set option.
*/
static boolean
_p_deep( struct inc_args *conf, boolean brace, char *start, int size )
{
    char buf[OPTS_MAX_SIZE];
 
    if( brace == FALSE ) AERROR("argument 'deep()' needs one option", FALSE);   
 
    buf[0] = '\0';
    strncat(buf, start, size);   
    conf->deep = atoi(buf);
    if( conf->deep < 0 || conf->deep >= 256 )
        AERROR("deep() option should be 0 or bigger, and smaller than 256", FALSE);
    
    return TRUE;
}



/*
** Switch ON cur and set options.
*/
static boolean
_p_cur( struct inc_args *conf, boolean brace, char *start, int size )
{
    if( brace == FALSE ) AERROR("argument 'cur()' need options", FALSE);

    conf->cur = TRUE;
    if( _p_fill_perms(&conf->_cur, start, size) == FALSE )
    {
        conf->cur = FALSE;
        return FALSE;
    }

    return TRUE;
}



/*
** Switch ON sub and set options.
*/
static boolean
_p_sub( struct inc_args *conf, boolean brace, char *start, int size )
{
    if( brace == FALSE ) AERROR("argument 'sub()' need options", FALSE);

    conf->sub = TRUE;
    if( _p_fill_perms(&conf->_sub, start, size) == FALSE )
    {
        conf->sub = FALSE;
        return FALSE;
    }

    return TRUE;
}



/*
** Free memory from VAR. 
*/
static void
p_free_var( VAR *root )
{
    VAR	  *tmp = root;


#ifndef NO_DEBUG
    printf("\n*-*-*-*-*-*-*-*-*-*-\n"\
           "| Freeing memory from variables.\n"\
           "-*-*-*-*-*-*-*-*-*-*-\n");
#endif    

    while( tmp != NULL ) 
    {
	root = tmp;
	tmp = tmp->next;

#ifndef NO_DEBUG
	printf("Variable:|%s|\n",root->name);
	printf("Contents:|%s|\n\n",root->contents);
#endif

	free(root->name);
	free(root->contents);
	free(root);
    }
}



/*
** Free memory from INCLUDE.
*/
void
p_free( INCLUDE *inc_root )
{
    INCLUDE	*config_tmp = inc_root;
    EXCLUDE	*exc_temp;
    FILES	*inc_temp;


#ifndef NO_DEBUG
    printf("\n*-*-*-*-*-*-*-*-*-*-\n"\
           "| Freeing memory from config.\n"\
           "-*-*-*-*-*-*-*-*-*-*-\n");
#endif    
    

    /* Free memory from config. */
    while( config_tmp != NULL ) 
    {
	inc_root = config_tmp;
	config_tmp = config_tmp->next;

#ifndef NO_DEBUG
	printf("Line:%d\n* Include:\ncheck:%d change:%d deep:%d cur:%d sub:%d\n",\
	        inc_root->line_in_config, inc_root->inc_args.check, inc_root->inc_args.change, inc_root->inc_args.deep,
	        inc_root->inc_args.cur, inc_root->inc_args.sub);
	if( inc_root->inc_args.sub == TRUE ) 
	    printf("Sub perms:%o user:|%s| group:|%s|\n",\
	            inc_root->inc_args._sub.mode, inc_root->inc_args._sub.user, inc_root->inc_args._sub.group);
	if( inc_root->inc_args.cur == TRUE ) 
	    printf("Cur perms:%o user:|%s| group:|%s|\n",\
                    inc_root->inc_args._cur.mode, inc_root->inc_args._cur.user, inc_root->inc_args._cur.group);

        printf("Files:");
#endif

        /* Free INCLUDE files list. */
	while( inc_root->include != NULL ) 
	{
	    inc_temp = inc_root->include;

#ifndef NO_DEBUG
	    printf("[%s]",inc_temp->filename);
#endif

	    inc_root->include = inc_root->include->next;
	    free(inc_temp->filename);
	    free(inc_temp);
	}


#ifndef NO_DEBUG
        printf("\n* Exclude:\n");
#endif
        /* Free exclude files list. */
        while( inc_root->exclude != NULL )
        {
	    exc_temp = inc_root->exclude;

#ifndef NO_DEBUG
            printf("Sub:%d Cur:%d Deep:%d File:|%s|\n",exc_temp->exc_args.sub, exc_temp->exc_args.cur, exc_temp->exc_args.deep, exc_temp->filename);
#endif

	    inc_root->exclude = inc_root->exclude->next;
	    free(exc_temp);
        }

        /* Free node. */
	if( inc_root->inc_args.sub == TRUE ) { free(inc_root->inc_args._sub.user); free(inc_root->inc_args._sub.group); }	
	if( inc_root->inc_args.cur == TRUE ) { free(inc_root->inc_args._cur.user); free(inc_root->inc_args._cur.group); }
	free(inc_root);

#ifndef NO_DEBUG
        printf("\n\n");
#endif
    }
}
