/*
 * apGuard - Access Permissions Guard.
 * Copyright (C) 2005-2006 Shobbo, urug
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <unistd.h>    /* chown */
#include <stdarg.h>    /* va_start */
#include <stdio.h>     /* fprintf */
#include <string.h>    /* strerror */
#include <sys/types.h> /* chmod */
#include <sys/stat.h>  /* chmod */
#include <time.h>      /* localtime */
#include <dirent.h>    /* chdir */
#include <errno.h>     /* errno */
#include <pwd.h>       /* getpwnam */
#include <grp.h>       /* getgrnam */
#include "parser.h"
#include "perms.h"


int			cur_line;	/* Current rule line. */
COMMAND			*gcmd;		/* Command line args. */
char			*path[256];	/* Path to file or directory (sub) */

static struct mode
{
    uid_t	uid;
    gid_t	gid;
    mode_t	mode;
} _cur, _sub;



/*
** Print info to output, or write to logfile. Last argument must be NULL.
** Space is only after first argument. Use GET_PATH to attach all content
** of path.
*/
static void
error( int get_path_pos, ... )
{
    FILE	*clf = NULL;
    boolean	quiet = gcmd->quiet;
    char	*el, cur_time[18], **_path;
    int		counter;
    va_list	ap;
    time_t	t;
    struct tm	*tm;


    /* initialize ap. */
    va_start( ap, get_path_pos );

    /* Open logfile, write current time to cur_time. */
    if( gcmd->log_file != NULL && (clf = fopen(gcmd->log_file, "a")) != NULL )
    {
        t = time(NULL);
        if( (tm = localtime(&t)) != NULL )
            strftime(cur_time, sizeof(cur_time), "%x %X",tm);
    }
    else if( gcmd->log_file != NULL )
        perror(gcmd->log_file);

    /* Print line and time. */
    if( quiet == FALSE ) fprintf(stdout, "Line(%d) ", cur_line);
    if( clf != NULL )    fprintf(clf, "%s line(%d) ", cur_time, cur_line);

    /* Print all strings. */
    for( counter = 1; (el = va_arg(ap, char *)) != NULL; counter++ )
    {
        if( counter == get_path_pos )
            for( _path = path; *_path != NULL; _path++ )
            {
                if( quiet == FALSE )  fprintf(stdout, "%s/", *_path );
                if( clf != NULL )     fprintf(clf, "%s/", *_path );
            }
   
        if( quiet == FALSE )  fprintf(stdout, "%s", el);
        if( clf != NULL )     fprintf(clf, "%s", el);
    }

    /* end. */
    va_end( ap );

    /* Write new line, close stream. */
    if( quiet == FALSE ) fprintf(stdout, "\n");
    if( clf != NULL ) {  fprintf(clf, "\n"); fclose(clf); }
}



/*
** Save uid, gid and mode in _cur and _sub.
*/
static boolean
save_uidgid( INCLUDE *inc )
{
    struct passwd	*pass;
    struct group	*grp;
    struct inc_perms	*src[2];
    struct mode		*dst[2];
    short 		i, opt[2];


    opt[0] = inc->inc_args.sub;
    src[0] = &inc->inc_args._sub;
    dst[0] = &_sub;
    opt[1] = inc->inc_args.cur;
    src[1] = &inc->inc_args._cur;
    dst[1] = &_cur;

    /* Save gid, uid and mode for from INCLUDE in _cur and _sub */
    for( i = 0 ; i < 2 ; i++ )
    {
        if( opt[i] == FALSE ) continue;

	/* Save file permissions to our local struct. */
	dst[i]->mode = src[i]->mode;

	/* Get uid from username. */
	if( (pass = getpwnam(src[i]->user)) == NULL )
	{
	    error(-1, "unknown owner name: ", src[i]->user, NULL);
	    return FALSE;
        }
        dst[i]->uid = pass->pw_uid;

	/* Save gid. */
	if( (grp = getgrnam(src[i]->group)) == NULL )
	{
	    error(-1, "unknown group name: ", src[i]->group, NULL);
            return FALSE;
        }
        dst[i]->gid = grp->gr_gid;
    }

    return TRUE;
}



/* 
** Function checks, and saves current file status.
*/
static boolean 
get_current_status( char *f, struct mode *fm )
{
    struct stat ss;


    /* Get file status. */
    if( lstat(f, &ss) == -1 )
    {
        error(-1, strerror(errno), ": ", f, NULL);
        return FALSE;
    }

    /* Save into fm current file modes. */
    fm->mode = ss.st_mode;
    fm->uid = ss.st_uid;
    fm->gid = ss.st_gid;

    return TRUE;
}



/*
** Argument 'change'
*/
static void 
change( char *f, struct mode *good_status, struct mode *cur_status )
{
    mode_t	cur_mode;
	

    /* Skip symbolic links. */
    if( S_ISLNK(cur_status->mode) )
        return;

    /* Write mode. */
    cur_mode = cur_status->mode & 8191; /* 0|000|111|111|111|111 */

    /* First check owner and group. Why first ? Because when owner and group
       are changed, suid/sgid dissapear. */
    if( (cur_status->uid != good_status->uid) || (cur_status->gid != good_status->gid) )
        if( chown(f, good_status->uid, good_status->gid) == -1 )
    	    error(-1, "chown: ", strerror(errno), ": ", f, NULL);

    /* Check mode. */
    if( cur_mode != good_status->mode )
        if( chmod(f, good_status->mode) == -1 )
            error(-1, "chmod: ", strerror(errno), ": ", f, NULL);
}



/*
** Argument 'check'
*/
static void 
check( char *f, struct mode *good_status, struct mode *cur_status )
{
    mode_t 		cur_mode;
    char		buf[3] = { '\0', '\0', '\0' };
    unsigned char	flags = 0;


    /* Skip symbolic links. */
    if( S_ISLNK(cur_status->mode) )
        return;

    /* Save current permissions. */
    cur_mode = cur_status->mode & 8191; /* 0|000|111|111|111|111 */

    /* Compare current status with good status from config file. */
    if( cur_mode != good_status->mode ) flags |= 1;
    if( cur_status->uid != good_status->uid ) flags |= 2;
    if( cur_status->gid != good_status->gid ) flags |= 4;

    /* Check. */
    if( flags )
    {
        if( flags & 1 ) strcat(buf, "M");
        if( flags & 2 ) strcat(buf, "O");
        if( flags & 4 ) strcat(buf, "G");

        /* Print full path. */
        error(5, "invalid status ", "(", buf, "): ", f, NULL);
    }
}
  


/*
** Argument 'cur()'
*/
static void 
cur( INCLUDE *inc, struct mode *cur_status, struct mode *good_status, char *f )
{
    /* Good place to use debug. */
    if( gcmd->debug == TRUE )
        error(2, "current: ", f, NULL );

    /* check. */
    if( inc->inc_args.check == TRUE )
        check(f, good_status, cur_status );

    /* change */    
    if( inc->inc_args.change == TRUE )
        change(f, good_status, cur_status );
}



/*
** Check if file or directory is excluded.
** TRUE: excluded, FALSE: no.
*/
static boolean
check_exclude( EXCLUDE *exclude )
{
    EXCLUDE	*current;
    char	*str;
    int		i, path_size, str_size, cmp_size;
    int		cur_deep, str_deep, size;
    boolean	sub, cur;
    int 	deep;

    
    if( exclude == NULL )
        return FALSE;

    /* Check all elements from exclude list. */
    for( current = exclude; current != NULL ; current = current->next )
    {
        /* save current exclude params. */
        cur = current->exc_args.cur;
        sub = current->exc_args.sub;
        deep = current->exc_args.deep;

        /* Check if current filename should be excluded. */
        str = current->filename;
        str_size = size = strlen(str);
        for( i = 0; str_size > 0 && str < current->filename + size && path[i] != NULL ; i++ )
        {
            /* Compare with size of path element or filename. */
            path_size = strlen(path[i]);
            if( path_size <= str_size )
                cmp_size = path_size;
            else
                cmp_size = str_size;

            /* Compare */
            if( strncmp(str, path[i], cmp_size) == 0 )
            {
                /* Save new string size (Current size - size of compared element + 1, 
                   because I skip '/' ) */
                str_size -= cmp_size + 1;

                /* If size is smaller than 0, parse arguments. */
                if( str_size <= 0 )
                {
                    i++; /* Skip current element in path. */

                    /* If cur == TRUE, and if path is equal to current file/directory return 
                       true. If user declared sub, count current deep. */
                    if( path[i] == NULL )
                    {
                        if( cur == TRUE ) 
                            return TRUE;
                    }
                    else if( sub == TRUE )
                    {
                        /* Deep 0 match all subdirectories. */
                        if( deep == 0 )
                            return TRUE;
                
                        /* Count string deep. */
                        for( str_deep = 1, str = current->filename; *str != '\0';  str++ )
                            if( *str == '/') str_deep++;

                        /* Count current deep (One dir, one element of array). */
                        for( cur_deep = i = 0; path[i] != NULL; i++ )
                            for( cur_deep++, str = path[i] ; *str != '\0' ; str++ )
                                if( *str == '/' )
                                    cur_deep++;
                
                        /* Check deep */
                        if( cur_deep - str_deep <= deep )
                            return TRUE;
                        else
                            break;
                    }
            
                    break;
                }

                /* Save string position after '/' sign. */
                str = &str[cmp_size] + 1;
            }
            else
                break;
                
        }
    }

    return FALSE;
}



/*
** Argument 'sub()'
*/
static void 
sub( INCLUDE *inc, char *d )
{
    DIR 		*dp;
    struct dirent	*dir_cont;
    struct mode		current_status;
    static unsigned int cd = 1;
    static unsigned int _ph;
    boolean		status;


    /* Dont go deeper than user need. */
    if( inc->inc_args.deep != 0 && cd > inc->inc_args.deep )
        return;

    /* Dont allow to go deepest than 256 */
    if( _ph >= 256-1 )
    {
        error(-1, "can't go deepest. ", NULL);
        return;
    }

    /* Open directory. */
    if( (dp = opendir(d)) == NULL )
    {
        error(-1, "can't open directory: ", d, NULL);
        return;
    }

    /* Go into that directory. */
    if( chdir(d) == 0 )
        cd++;
    else 
    {
        error(-1, "can't go into directory: ", d, NULL);
        closedir(dp);
        return;
    }

    /* Write current directory. */
    path[_ph++] = d;
    path[_ph] = NULL;	/* NULL should be always at the end. */

    /* Check all files and directiories in path. */
    while( (dir_cont = readdir(dp)) != NULL )
    {
        if( strcmp("..", dir_cont->d_name) == 0 || strcmp(".", dir_cont->d_name) == 0 )
            continue;

        /* Exclude one element or whole directory. */
        path[_ph] = dir_cont->d_name;
        status = check_exclude( inc->exclude );
        path[_ph] = NULL;

        /* Get current status of file/directory. */
        if( get_current_status(dir_cont->d_name, &current_status) == -1 )
            continue;

        /* If current dir_cont is a directory, go deeper. If not, make changes. */
        if( S_ISDIR(current_status.mode) ) 
        {
            /* Check current directory and go deeper. */
            if( status == FALSE ) cur(inc, &current_status, &_sub, dir_cont->d_name );
            sub(inc, dir_cont->d_name);
        }
        else if( status == FALSE )
            cur(inc, &current_status, &_sub, dir_cont->d_name );
    }

    path[_ph--] = NULL;	/* Write NULL. */
    cd--;
    chdir("..");
    closedir(dp);
}



/*
** Run all check functions on filename.
*/
static void 
do_job( INCLUDE *inc, char *f )
{
    struct mode	fm;


    /* Get current file status. */
    if( get_current_status(f, &fm) == FALSE )
        return;

    /* Clear path. */
    path[0] = NULL;

    /* Check current file/dir. */ 
    if( inc->inc_args.cur == TRUE )
        cur(inc, &fm, &_cur, f);

    /* Check subdirectories. */
    if( inc->inc_args.sub == TRUE )
    {
        /* Print error if got file, not directory. */
        if( S_ISDIR(fm.mode) )
            sub(inc, f);
        else
        {
            error(-1, "expected directory, got file: ", f, NULL );
            return;
        }
    }
}



/*
** Start.
*/
void
mod_perms( INCLUDE *inc_root, COMMAND *cmd )
{
    INCLUDE	*r;
    FILES	*fs;

    
    /* Copy command line args to global variable. */
    gcmd = cmd;

    /* Start checking files. */
    for( r = inc_root ; r != NULL ; r = r->next )
    {
    	/* Save current line */
    	cur_line = r->line_in_config;

    	/* Save uid, gid and mode into _cur and _sub. */
        if( save_uidgid(r) == FALSE )
            continue;

        /* Check all filenames in current rule. */
        for( fs = r->include; fs != NULL; fs = fs->next )
            do_job(r, fs->filename);
    }
}
